uniform sampler2D bgl_DepthTexture;
uniform sampler2D bgl_RenderedTexture;
uniform float bgl_RenderedTextureWidth;
uniform float bgl_RenderedTextureHeight;
uniform sampler2D iChannel0;
uniform sampler2D bgl_DataTextures[7];







//THIS NEEDS TO MATCH YOUR CAMERA SETTINGS---------------------
const float znear = 0.1;    //Z-near
const float zfar = 100.0;   //Z-far
const float fov = 50.0;     //FoV
//-------------------------------------------------------------


//USER VARIABLES-----------------------------------------------
const float intensity = 0.5;            //Intensity of the AO effect
const float sampleRadius = 2.0;         //Radius of the AO, bigger values need more performance

const float sampleDirections = 8.0;    //Main sample count, affects performance heavily 
const float sampleSteps = 8.0;         //SubSample count for path walking, affects performance


const bool useAttenuation = true;       //Applies attenuation to each AO sample, true = better quality, off = better performance
const float attenuationScale = 0.3;     //Depth scale of the attenuation, different values look better depending on your scene


const float angleBias = 0.05;            //Brightens up the AO effect. Higer values mean less noise but less accuracy


const bool noise = true;                //Use noise instead of pattern for sample dithering, much slower but EXTREMLY better looking
const float noiseamount = 0.8;          //Per-Pixel noise amount, bigger values need more performance
const float jitterAmount = 0.8;         //Per-Sample noise amount, bigger values need more performance


const float lumInfluence = 0.5;         //Influence of the luminance on the AO effect
const vec3 aoColor = vec3(0.264, 0.531, 1.0) * 0.4; // Color of the AO effect


const bool onlyAO = true;                //Only show AO pass for debugging
const bool externalBlur = false;         //Store AO in alpha pass for a later blur

const float fogCutoff = 50.0;            // distance where AOu cuts off
//-------------------------------------------------------------












const float width = bgl_RenderedTextureWidth;
const float height = bgl_RenderedTextureHeight;


vec2 texCoord = gl_TexCoord[0].st;


const float aspectratio = width/height;
const float thfov = tan(fov * 0.0087266462597222);
const float TWO_PI = 6.283185307;


vec3 getLinearColor(vec2 coord)
{    
    vec3 C = vec3(texture2D(bgl_RenderedTexture, coord));
    C.r = pow(C.r, 2.2);
    C.g = pow(C.g, 2.2);
    C.b = pow(C.b, 2.2);
    return C.rgb;
}


vec3 sRGBToLinear(vec3 C)
{
    C.r = pow(C.r, 2.2);
    C.g = pow(C.g, 2.2);
    C.b = pow(C.b, 2.2);
    return C.rgb;
}


vec3 linearTosRGB(vec3 C)
{
    C.r = pow(C.r, 0.45454545);
    C.g = pow(C.g, 0.45454545);
    C.b = pow(C.b, 0.45454545);
    return C.rgb;
}


float getLinearDepth(vec2 coord)
{
    float zdepth = texture2D(bgl_DepthTexture, coord).x;
    return zfar*znear / (zfar + zdepth * (znear - zfar));
}


vec3 getViewVector(vec2 coord)
{
    vec2 ndc = (coord * 2.0 - 1.0);
    return vec3(ndc.x*thfov, ndc.y*thfov/aspectratio, 1.0);
}


vec3 getViewPosition(vec2 coord)
{
    return getViewVector(coord) * getLinearDepth(coord);
}



const float pW = 1.0/width;
const float pH = 1.0/height;

vec3 getViewNormal(vec2 coord)
{
    return normalize((texture2D(bgl_DataTextures[0], coord).rgb - vec3(0.5)) * vec3(-2.0, -2.0, 2.0));
}


const float rand(vec2 co) {
    if (noise) {
        return fract(sin(dot(co.xy,vec2(12.9898,78.233)*4.0)) * 43758.5453)*2.0-1.0;
    } else {
        return ((fract(1.0-co.s*(width/2.0))*0.3)+(fract(co.t*(height/2.0))*0.6))*2.0-1.0;
    }
}


const vec2 noiseOffset2d = vec2(max(1.0 - noiseamount, 0.0), 0.0);

vec2 rand2D(vec2 coord) {
    float noiseX = ((fract(1.0-coord.s*(width/2.0))*0.25)+(fract(coord.t*(height/2.0))*0.75))*2.0-1.0;
    float noiseY = ((fract(1.0-coord.s*(width/2.0))*0.75)+(fract(coord.t*(height/2.0))*0.25))*2.0-1.0;
    
    if (noise)
    {
        noiseX = clamp(fract(sin(dot(coord ,vec2(12.9898,78.233))) * 43758.5453),0.0,1.0)*2.0-1.0;
        noiseY = clamp(fract(sin(dot(coord ,vec2(12.9898,78.233)*2.0)) * 43758.5453),0.0,1.0)*2.0-1.0;
    }
    return (normalize(vec2(noiseX,noiseY)) * noiseamount) + noiseOffset2d;
}


const mat4 getViewProjectionMatrix() {
    mat4 result;
    
    float frustumDepth = zfar - znear;
    float oneOverDepth = 1 / frustumDepth;


    result[0][0] = 1 / thfov;
    result[1][1] = aspectratio * result[0][0];
    result[2][2] = zfar * oneOverDepth;
    result[2][3] = 1;
    result[3][2] = (-zfar * znear) * oneOverDepth;
    
    return result;
}


vec2 ComputeFOVProjection(vec3 pos, mat4 VPM)
{
    vec4 offset = vec4(pos, 1.0);
    offset = VPM * offset;
    offset.xy /= offset.w;
    return offset.xy * 0.5 + 0.5;
}

const mat4 VPM = getViewProjectionMatrix();
const mat4 viewProjectionInverseMatrix  = inverse(VPM);
    
const float radiusSS = sampleRadius;
const vec4 temp0 = viewProjectionInverseMatrix * vec4(0.0, 0.0, -1.0, 1.0);
const vec4 temp1 = viewProjectionInverseMatrix * vec4(radiusSS, 0.0, -1.0, 1.0);
const float outDist = length(temp1.xyz - temp0.xyz);
const float fovFact = tan(radiusSS * fov * 0.0087266462597222);
    
const float theta = TWO_PI / float(sampleDirections);
const float cosTheta = cos(theta);
const float sinTheta = sin(theta);
const mat2 deltaRotationMatrix = mat2(cosTheta, -sinTheta, sinTheta, cosTheta);

const vec2 sampleNoise = normalize(rand2D(texCoord));
const mat2 rotationMatrix = mat2(sampleNoise.x, -sampleNoise.y, 
                               sampleNoise.y,  sampleNoise.x);

void main()
{
    
    vec3 color = texture2D(bgl_RenderedTexture,texCoord).rgb;
    
    vec3 originVS = getViewPosition(texCoord);
    vec3 normalVS = getViewNormal(texCoord);
    
    if(originVS.z > fogCutoff) {
        gl_FragColor.rgb = color;
        gl_FragColor.a = 1.0;
        
        if(onlyAO) {
            gl_FragColor.rgb = vec3(1.0);
            gl_FragColor.a = 1.0;
        }
        
        return;
    }
    
    const float radiusWS = min(tan(radiusSS * fov * 0.0087266462597222) * originVS.z, length(temp0.xyz - temp1.xyz));
    vec2 deltaUV = vec2(1.0, 0.0) * (radiusSS / (float(sampleDirections * sampleSteps) + 1.0));

    deltaUV = rotationMatrix * deltaUV;

    
    float occlusion = 0.0;
    float jitter = rand(texCoord) * jitterAmount;
    
    for(int i = 0; i < sampleDirections; i++) {
        deltaUV = deltaRotationMatrix * deltaUV;
        
        vec2 sampleDirUV = deltaUV / originVS.z * vec2(20.0);
        float oldAngle = angleBias;
        
        for(int j = 0; j < sampleSteps; j++) {
            vec2 offsetNoise = mod(jitter + vec2(j * 0.429, j * 0.555), 1.0);
            vec2 sampleUV = texCoord + (mix(vec2(1.0), offsetNoise, jitterAmount) * sampleDirUV);
            vec3 sampleVS = getViewPosition(sampleUV);
            vec3 sampleDirVS = (sampleVS - originVS);
            
            float gamma = 1.570796326 - acos(dot(normalVS, normalize(sampleDirVS)));
            
            if (gamma > oldAngle) {
                float value = sin(gamma) - sin(oldAngle);
                
                if(useAttenuation) {
                    float attenuation = clamp(1.0 - pow(length(sampleDirVS) / radiusWS * attenuationScale, 2.0), 0.0, 1.0);
                    occlusion += attenuation * value;
                } else {
                    occlusion += value;
                }
                
                oldAngle = gamma;
            }
        }
    }
    
    occlusion = 1.0 - occlusion / float(sampleDirections);
    occlusion = clamp(pow(occlusion, 1.0 + intensity), 0.0, 1.0);
    
    vec3 lumcoeff = vec3(0.299,0.587,0.114);
    float lum = dot(color.rgb, lumcoeff) * lumInfluence;
    
    occlusion = lum + ((1.0 - lum) * occlusion);
    
    const vec3 aoMul = mix(aoColor, vec3(1.0), occlusion);
    
    gl_FragColor.rgb = color * aoMul;
    gl_FragColor.a = occlusion;
    
    if(externalBlur) {
        gl_FragColor.rgb = color;
        gl_FragColor.a = occlusion;
    }
    
    if(onlyAO) {
        gl_FragColor.rgb = vec3(occlusion);
        gl_FragColor.a = 1.0;
    }
}



